"""
--- Day 15: Beacon Exclusion Zone ---
https://adventofcode.com/2022/day/15
"""
from itertools import combinations

from aocd import data
from intervaltree import IntervalTree
from parse import parse

from aoc_wim.zgrid import manhattan_distance


class Sensor:
    def __init__(self, z, b):
        self.z = z  # sensor position
        self.b = b  # nearest beacon position
        # radius just out of range - i.e. ball perimeter
        self.r = manhattan_distance(z - b) + 1
        # top, right, bottom, left corners of the sensor range
        self.corner = [self.z + self.r * dz for dz in (-1j, 1, 1j, -1)]

    @classmethod
    def fromline(cls, line):
        x1, y1, x2, y2 = parse(template, line).fixed
        return cls(x1 + 1j*y1, x2 + 1j*y2)

    def __contains__(self, z):
        # is the point z detectable from this sensor
        return manhattan_distance(self.z - z) < self.r

    def range_on(self, y):
        # range of x coordinates that are visible at given y coordinate (if any)
        x0, y0 = int(self.z.real), int(self.z.imag)
        r0 = self.r - 1
        dy = abs(y0 - y)
        return range(x0 - r0 + dy, x0 + r0 - dy + 1)


template = "Sensor at x={:d}, y={:d}: closest beacon is at x={:d}, y={:d}"
sensors = [Sensor.fromline(x) for x in data.splitlines()]

y = 2000000
maxd = 4000000
if len(sensors) == 14:  # example data - smaller bounds are specified in the prose
    y = 10
    maxd = 20

tree = IntervalTree()
for sensor in sensors:
    if xs := sensor.range_on(y):
        tree[xs.start:xs.stop] = 1
tree.merge_overlaps()
a = sum(i.length() for i in tree)
# discount any beacons actually on the line
a -= len({int(s.b.real) for s in sensors if int(s.b.imag) == y})
print("answer_a:", a)


# the point we're looking for should be just beyond the perimeter of 4 different scanners
# the perimeter in manhattan distance is a diamond shape ◇
# when 4 edges intersect it will look like a rotated #
# we can find intersections of edges like / and \ use as candidates for the search
# when such a point is out of range of *all* the sensors, we have found the solution

def line_intersection(z1, z2, z3, z4):
    # point of intersection of two line segments, specified by 4 complex numbers
    # returns tuple of z_intersection (complex), on_segment1 (bool), on_segment2 (bool)
    z12 = z1 - z2
    z34 = z3 - z4
    if z12 == 0 or z34 == 0:
        raise ValueError("equal start/end points")
    denom = z12.real * z34.imag - z12.imag * z34.real
    if denom == 0:
        raise ValueError("lines are parallel")
    n1 = z1.real * z2.imag - z1.imag * z2.real
    n2 = z3.real * z4.imag - z3.imag * z4.real
    real = (n1 * z34.real - n2 * z12.real) / denom
    imag = (n1 * z34.imag - n2 * z12.imag) / denom
    z = real + 1j * imag
    t = (z1.real - z3.real) * (z3.imag - z4.imag) - (z1.imag - z3.imag) * (z3.real - z4.real)
    u = (z1.real - z3.real) * (z1.imag - z2.imag) - (z1.imag - z3.imag) * (z1.real - z2.real)
    return z, 0 <= t/denom <= 1, 0 <= u/denom <= 1


def tuning_frequency(z):
    return int(z.real) * 4000000 + int(z.imag)


def unique_point_out_of_range():
    for s1, s2 in combinations(sensors, 2):
        for i in range(4):
            z1, z2 = s1.corner[(i + 0) % 4], s1.corner[(i + 1) % 4]
            for j in [0, 2]:
                z3, z4 = s2.corner[(i + j + 1) % 4], s2.corner[(i + j + 2) % 4]
                z, in1, in2 = line_intersection(z1, z2, z3, z4)
                if in1 and in2 and z.real.is_integer() and z.imag.is_integer():
                    if not any(z in s for s in sensors):
                        if 0 <= z.real <= maxd and 0 <= z.imag <= maxd:
                            return z


z = unique_point_out_of_range()
b = tuning_frequency(z)
print("answer_b:", b)



"""
--- Day 24: Arithmetic Logic Unit ---
https://adventofcode.com/2021/day/24
"""
import parse
from aocd import data


class Comp:
    def __init__(self, code, inputs):
        self.lines = code.splitlines()
        self.reg = dict.fromkeys("wxyz", 0)
        if isinstance(inputs, int):
            inputs = [int(n) for n in str(inputs)]
        self.input_iter = iter(inputs)

    def inp(self, a):
        self.reg[a] = next(self.input_iter)

    def add(self, a, b):
        try:
            b = int(b)
        except ValueError:
            b = self.reg[b]
        self.reg[a] += b

    def mul(self, a, b):
        try:
            b = int(b)
        except ValueError:
            b = self.reg[b]
        self.reg[a] *= b

    def div(self, a, b):
        try:
            b = int(b)
        except ValueError:
            b = self.reg[b]
        self.reg[a] //= b

    def mod(self, a, b):
        try:
            b = int(b)
        except ValueError:
            b = self.reg[b]
        self.reg[a] %= b

    def eql(self, a, b):
        try:
            b = int(b)
        except ValueError:
            b = self.reg[b]
        self.reg[a] = int(self.reg[a] == b)

    def run(self):
        for i, line in enumerate(self.lines):
            op, *args = line.split()
            f = getattr(self, op)
            f(*args)


optimize = """\
inp w
mul x 0
add x z
mod x 26
div z {:d}
add x {:d}
eql x w
eql x 0
mul y 0
add y 25
mul y x
add y 1
mul z y
mul y 0
add y w
add y {:d}
mul y x
add z y"""


def solve(data, part="a"):
    div_x_y = [r.fixed for r in parse.findall(optimize, data)]
    if not div_x_y:
        return
    incs = [y for div, x, y in div_x_y if div == 1]
    n_high = int("9" * len(incs))
    n_low = int("1" * len(incs))
    ns = range(n_high, n_low - 1, -1)
    if part == "b":
        ns = reversed(ns)
    for n in ns:
        s_n = str(n)
        if "0" in s_n:
            continue
        it = (int(d) for d in s_n)
        inputs = []
        z = 0
        for div, x, y in div_x_y:
            if div == 1:
                inputs.append(next(it))
                z = 26 * z + inputs[-1] + y
            else:
                i = z % 26 + x
                if not 1 <= i <= 9:
                    break
                inputs.append(i)
                z //= 26
        if z == 0:
            result = sum(10**i * n for i, n in enumerate(reversed(inputs)))
            # verify the result actually works in the VM!
            comp = Comp(code=data, inputs=inputs)
            comp.run()
            assert (
                len(inputs) == 14
                and all(1 <= i <= 9 for i in inputs)
                and comp.reg["z"] == 0
            )
            return result


print("answer_a:", solve(data, part="a"))
print("answer_b:", solve(data, part="b"))



"""
--- Day 11: Corporate Policy ---
https://adventofcode.com/2015/day/11
"""
import re

from aocd import data


def req1(s):
    """
    Passwords must include one increasing straight of at least three letters, 
    like abc, bcd, cde, and so on, up to xyz. They cannot skip letters; abd 
    doesn't count.
    """
    for a, b, c in zip(s, s[1:], s[2:]):
        ord_a = ord(a)
        if (0, ord(b) % ord_a, ord(c) % ord_a) == (0, 1, 2):
            return True
    else:
        return False


def req2(s):
    """
    Passwords may not contain the letters i, o, or l, as these letters can be 
    mistaken for other characters and are therefore confusing.
    """
    return "i" not in s and "o" not in s and "l" not in s


def req3(s):
    """
    Passwords must contain at least two different, non-overlapping pairs of 
    letters, like aa, bb, or zz.
    """
    return len(re.findall(r"(.)\1", s)) >= 2


def is_valid(passwd):
    return req1(passwd) and req2(passwd) and req3(passwd)


def preprocess(s):
    result = ""
    for c in s:
        if c in "iol":
            result += {"i": "j", "o": "p", "l": "m"}[c]
            break
        else:
            result += c
    result += "a" * (8 - len(result))
    return result


def next_password(s):
    s = preprocess(s)
    alphabet = "abcdefghjkmnpqrstuvwxyz"
    tr = dict(zip(alphabet, alphabet[1:]))
    while True:
        s_list = list(s)
        i = -1
        while True:
            try:
                s_list[i] = tr[s_list[i]]
            except KeyError:
                s_list[i] = "a"
                i -= 1
            else:
                break
        s = "".join(s_list)
        if is_valid(s):
            return s


if __name__ == "__main__":
    first = next_password(data)
    print("answer_a:", first)
    print("answer_b:", next_password(first))



import logging
import re
from _md5 import md5
from functools import partial
from multiprocessing import current_process


log = logging.getLogger(__name__)
# logging.basicConfig(format="%(message)s", level=logging.INFO)


class AssembunnyComputer:
    def __init__(self, source, a0=0, b0=0, c0=0, d0=0):
        self.i = 0
        self.lines = source.splitlines()
        self.reg = {"a": a0, "b": b0, "c": c0, "d": d0}

    def run(self):
        while True:
            try:
                self.step()
            except IndexError:
                return

    def step(self):
        line = self.lines[self.i]
        log.debug("i=%02d %-8s %s", self.i, line, self.reg)
        op, *args = line.split()

        if op == "inc":
            asm = self.lines[self.i : self.i + 3]
            if len(asm) == 3 and asm[1].startswith("dec"):
                _, r_inc = asm[0].split()
                _, r_dec = asm[1].split()
                if asm[2].startswith("jnz"):
                    _, r_cond, j = asm[2].split()
                    if j == "-2" and r_cond == r_dec:
                        if self.reg[r_cond] > 0:
                            add = self.reg[r_cond]
                            msg = "patching out iadd loop (lines %d-%d): %s += %d"
                            log.info(msg, self.i, self.i + 2, r_inc, add)
                            self.reg[r_inc] += add
                            self.reg[r_cond] = 0
                            self.i += 3
                            return

        if op == "cpy" and self.lines[self.i : self.i + 6] == [
            "cpy b c",
            "inc a",
            "dec c",
            "jnz c -2",
            "dec d",
            "jnz d -5",
        ]:
            msg = "patching out multiply double loop (lines %d-%d): a += b * d"
            log.info(msg, self.i, self.i + 6)
            self.reg["a"] += self.reg["b"] * self.reg["d"]
            self.reg["c"] = self.reg["d"] = 0
            self.i += 6
            return

        getattr(self, op)(*args)
        self.i += 1

    def cpy(self, x, y):
        self.reg[y] = self.reg[x] if x in self.reg else int(x)

    def jnz(self, x, y):
        x = self.reg[x] if x in self.reg else int(x)
        y = self.reg[y] if y in self.reg else int(y)
        if x:
            self.i += y - 1

    def dec(self, r):
        self.reg[r] -= 1

    def inc(self, r):
        self.reg[r] += 1

    def tgl(self, di):
        di = self.reg[di] if di in self.reg else int(di)
        i = self.i + di
        try:
            line = self.lines[i]
        except IndexError:
            # If an attempt is made to toggle an instruction
            # outside the program, nothing happens
            return
        op, *args = line.split()
        if len(args) == 1:
            # For one-argument instructions, inc becomes dec,
            # and all other one-argument instructions become inc.
            op = "dec" if op == "inc" else "inc"
        elif len(args) == 2:
            # For two-argument instructions, jnz becomes cpy,
            # and all other two-instructions become jnz.
            op = "cpy" if op == "jnz" else "jnz"
        self.lines[i] = " ".join([op, *args])


def md5_miner(data, i0, n, keep, stretch=None, verbose=False):
    pid = current_process().pid
    if verbose:
        print(f"MD5 miner start {i0=} {pid=}")
    h0 = md5(data.encode("ascii")).copy
    results = []
    for i in range(i0, i0 + n):
        hash_ = h0()
        hash_.update(b"%d" % i)
        result = hash_.hexdigest()
        if stretch is not None:
            result = result.encode()
            for _ in range(stretch - 1):
                result = md5(result).hexdigest().encode()
            result = result.decode()
        if keep(result):
            results.append((i, result))
    mined = len(results)
    if verbose:
        print(f"MD5 miner stop  {i0=} {pid=} {mined=}")
    return results


def _keep_q05(digest, difficulty=5):
    return digest.startswith("0" * difficulty)


def _keep_q14(digest):
    pat3 = re.compile(r"(.)\1{2}")
    return pat3.search(digest) is not None


md5_miner_q05 = partial(md5_miner, keep=_keep_q05)
md5_miner_q14a = partial(md5_miner, keep=_keep_q14)
md5_miner_q14b = partial(md5_miner, keep=_keep_q14, stretch=2017)



"""
--- Day 1: Calorie Counting ---
https://adventofcode.com/2022/day/1
"""
from heapq import nlargest

from aocd import data

cals = []
for chunk in data.split("\n\n"):
    cals.append(sum(map(int, chunk.split())))

print("answer_a:", max(cals))
print("answer_b:", sum(nlargest(3, cals)))



"""
--- Day 1: Report Repair ---
https://adventofcode.com/2020/day/1
"""
from collections import Counter

from aocd import data

from aoc_wim.aoc2020 import find_pair


counter = Counter(int(x) for x in data.splitlines())
x, y = find_pair(counter)
print(f"answer_a: {x} * {y} == {x * y}")

for z in list(counter):
    counter[z] -= 1
    try:
        x, y = find_pair(counter, target=2020 - z)
    except TypeError:
        counter[z] += 1
    else:
        assert x + y + z == 2020
        print(f"answer_b: {x} * {y} * {z} == {x * y * z}")
        break



"""
--- Day 19: A Series of Tubes ---
https://adventofcode.com/2017/day/19
"""
from aocd import data

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
z = data.splitlines()[0].index("|")
assert grid[z] == "|"
dz = ZGrid.down
letters = ""
n_steps = 0
while True:
    z += dz
    n_steps += 1
    a = grid.get(z, " ")
    if a == "+":
        dz *= ZGrid.turn_left
        if grid.get(z + dz, " ") == " ":
            dz *= ZGrid.turn_around
    elif a not in "-| ":
        letters += a
    elif a == " ":
        break

print("answer_a:", letters)
print("answer_b:", n_steps)



"""
--- Day 17: No Such Thing as Too Much ---
https://adventofcode.com/2015/day/17
"""
from aocd import data

from aoc_wim.stuff import subset_sum

vals = [int(n) for n in data.splitlines()]
liters = 25 if len(vals) == 5 else 150
subsets = list(subset_sum(vals, liters))
print("answer_a:", len(subsets))
fewest = min(subsets, key=len)
print("answer_b:", sum(len(s) == len(fewest) for s in subsets))



"""
--- Day 5: Binary Boarding ---
https://adventofcode.com/2020/day/5
"""
from aocd import data

tr = str.maketrans("FBLR", "0101")
ids = {int(x.translate(tr), 2) for x in data.splitlines()}
print("answer_a:", max(ids))
missing = set(range(min(ids), max(ids))) - ids
print("answer_b:", *missing)



"""
--- Day 22: Grid Computing ---
https://adventofcode.com/2016/day/22
"""
import numpy as np
from aocd import data


def parsed(data):
    lines = [x for x in data.splitlines() if x.startswith("/dev")]
    x, y = lines[-1].split()[0].replace("/dev/grid/node-x", "").split("-y")
    shape = int(y) + 1, int(x) + 1
    used = np.zeros(shape, dtype=int)
    avail = np.zeros(shape, dtype=int)
    for line in lines:
        node, _, used_, avail_, _ = line.split()
        node = node.replace("/dev/grid/node-x", "")
        col, row = [int(n) for n in node.split("-y")]
        used[row, col] = int(used_.rstrip("T"))
        avail[row, col] = int(avail_.rstrip("T"))
    return used, avail


used, avail = parsed(data)
part_a = sum((n <= avail).sum() - 1 for n in used.ravel())

used = (used / np.median(used)).round().astype(int)
used[used > 1] = 2
[row0], [col0] = np.where(used == 0)
h, w = used.shape
goal_row, goal_col = 0, w - 1
# manhattan distance between free space and goal data
dist_0_goal = row0 + (goal_col - col0)
has_detour = (used[:row0, col0] == 2).any()
w_min = np.where(used == 2)[1].min()  # wall end
# extra steps to go around the wall
detour = (col0 - (w_min - 1)) * 2 if has_detour else 0
# -1 because in moving free space to goal position, G is offset
move_data = 5 * (goal_col - 1)
part_b = dist_0_goal + detour + move_data

print("answer_a:", part_a)
print("answer_b:", part_b)



"""
--- Day 19: Not Enough Minerals ---
https://adventofcode.com/2022/day/19
"""
import math
import re

from aocd import data


lines = data.split("\n\n") if "\n\n" in data else data.splitlines()
bps = []
for line in lines:
    pk, ore3, ore2, ore1, clay, ore0, obs = map(int, re.findall(r"\d+", line))
    bps.append((pk, ore3, ore2, ore1, clay, ore0, obs, max(ore3, ore2, ore1, ore0)))


def search(stack, blueprint):
    _pk, ore3, ore2, ore1, clay, ore0, obs, ore = blueprint
    tri = [t * (t - 1) // 2 for t in range(33)]
    max_geodes = 0
    while stack:
        r, n0, n1, n2, n3, r0, r1, r2, r3, t = stack.pop()
        g = n0 + r0 * t
        if g > max_geodes:
            max_geodes = g
        if t == 1:
            continue  # on the very last minute, building another robot is not useful
        if r == 0 and not r1:
            continue  # you need at least one obsidian robot before considering a geode robot
        elif r == 1 and not r2:
            continue  # you need at least one clay robot before considering an obsidian robot
        if r == 1 and r1 >= obs:
            continue  # don't need any more obsidian than you can actually spend
        elif r == 2 and r2 >= clay:
            continue  # don't need any more clay than you can actually spend
        elif r == 3 and r3 >= ore:
            continue  # don't need any more ore than you can actually spend
        if n0 + t * r0 + tri[t] <= max_geodes:
            continue  # prune this branch - it can't beat our best so far
        if r == 0 and n1 >= obs and n3 >= ore0:  # get a geode robot
            stack += [(r, n0 + r0, n1 + r1 - obs, n2 + r2, n3 + r3 - ore0, r0 + 1, r1, r2, r3, t - 1) for r in range(4)]
        elif r == 1 and n2 >= clay and n3 >= ore1:  # get an obsidian robot
            stack += [(r, n0 + r0, n1 + r1, n2 + r2 - clay, n3 + r3 - ore1, r0, r1 + 1, r2, r3, t - 1) for r in range(4)]
        elif r == 2 and n3 >= ore2:  # get a clay robot
            stack += [(r, n0 + r0, n1 + r1, n2 + r2, n3 + r3 - ore2, r0, r1, r2 + 1, r3, t - 1) for r in range(4)]
        elif r == 3 and n3 >= ore3:  # get an ore robot
            stack += [(r, n0 + r0, n1 + r1, n2 + r2, n3 + r3 - ore3, r0, r1, r2, r3 + 1, t - 1) for r in range(4)]
        else:  # just wait - accumulates more minerals
            stack.append((r, n0 + r0, n1 + r1, n2 + r2, n3 + r3, r0, r1, r2, r3, t - 1))
    print(f"{max_geodes=} {blueprint=}")
    return max_geodes


s0 = 0, 0, 0, 0, 0, 0, 0, 1
print("answer_a:", sum(bp[0] * search([(2, *s0, 24), (3, *s0, 24)], bp) for bp in bps))
print("answer_b:", math.prod(search([(2, *s0, 32), (3, *s0, 32)], bp) for bp in bps[:3]))



"""
--- Day 13: Transparent Origami ---
https://adventofcode.com/2021/day/13
"""
from ast import literal_eval

from aocd import data

from aoc_wim.ocr import AOCR

zs, folds = data.replace(",", "+").split("\n\n")
d = {literal_eval(z + "j"): "#" for z in zs.split()}

for i, fold in enumerate(folds.splitlines()):
    axis, n = fold.split("=")
    n = int(n)
    for z in [*d]:
        if axis[-1] == "y" and z.imag > n:
            z_ = complex(z.real, 2 * n - z.imag)
        elif axis[-1] == "x" and z.real > n:
            z_ = complex(2 * n - z.real, z.imag)
        else:
            continue
        d[z_] = "#"
        del d[z]
    if i == 0:
        a = len(d)

print("answer_a:", a)
print("answer_b:", AOCR[d])



"""
--- Day 12: Digital Plumber ---
https://adventofcode.com/2017/day/12
"""
from aocd import data


d = {}
for line in data.replace(" <->", ",").splitlines():
    xs = [int(x) for x in line.split(", ")]
    d[tuple(xs)] = set(xs)

while True:
    for tuple_, set1 in d.items():
        try:
            match = next(k for k, set2 in d.items() if k != tuple_ and set1 & set2)
        except StopIteration:
            continue  # no match for this key - keep looking
        else:
            d[tuple_] = set1 | d.pop(match)
            break  # merged match and set1
    else:
        break  # no match for any key - we are done!

[set0] = [v for v in d.values() if 0 in v]
print("answer_a:", len(set0))
print("answer_b:", len(d))



"""
--- Day 22: Reactor Reboot ---
https://adventofcode.com/2021/day/22
"""
from dataclasses import dataclass

from aocd import data
from parse import parse


@dataclass
class Cuboid:
    # the non-empty 3D space of {x,y,z} with
    #   x0 <= x < x1
    #   y0 <= y < y1
    #   z0 <= z < z1

    val: str
    x0: int
    x1: int
    y0: int
    y1: int
    z0: int
    z1: int
    small: bool = False

    @classmethod
    def from_line(cls, line):
        template = "{} x={:d}..{:d},y={:d}..{:d},z={:d}..{:d}"
        val, x0, x1, y0, y1, z0, z1 = parse(template, line).fixed
        small = all(abs(v) <= 50 for v in (x0, x1, y0, y1, z0, z1))
        return cls(val, x0, x1 + 1, y0, y1 + 1, z0, z1 + 1, small)

    def __contains__(self, other):
        if not isinstance(other, Cuboid):
            return NotImplemented
        return (
            self.x0 <= other.x0 < other.x1 <= self.x1
            and self.y0 <= other.y0 < other.y1 <= self.y1
            and self.z0 <= other.z0 < other.z1 <= self.z1
        )


def total_volume(cuboids):
    xs = sorted(set([c.x0 for c in cuboids] + [c.x1 for c in cuboids]))
    ys = sorted(set([c.y0 for c in cuboids] + [c.y1 for c in cuboids]))
    zs = sorted(set([c.z0 for c in cuboids] + [c.z1 for c in cuboids]))
    result = 0
    for x0, x1 in zip(xs, xs[1:]):
        for y0, y1 in zip(ys, ys[1:]):
            for z0, z1 in zip(zs, zs[1:]):
                cell = Cuboid("", x0, x1, y0, y1, z0, z1)
                for cuboid in reversed(cuboids):
                    if cell in cuboid:
                        if cuboid.val == "on":
                            result += (x1 - x0) * (y1 - y0) * (z1 - z0)
                        break
    return result


cuboids = [Cuboid.from_line(x) for x in data.splitlines()]
small_cuboids = [c for c in cuboids if c.small]

print("answer_a:", total_volume(small_cuboids))
print("answer_b:", total_volume(cuboids))



"""
--- Day 1: Inverse Captcha ---
https://adventofcode.com/2017/day/1
"""
import numpy as np
from aocd import data


def part_a(data, roll=1):
    a = np.fromiter(data, dtype=int)
    b = np.roll(a, roll)
    return a[a == b].sum()


def part_b(data):
    return part_a(data, roll=len(data) // 2)


print("answer_a:", part_a(data))
print("answer_b:", part_b(data))



"""
--- Day 19: Medicine for Rudolph ---
https://adventofcode.com/2015/day/19
"""
from collections import defaultdict

from aocd import data


reactions, sep, element = data.partition("\n\n")
tr = defaultdict(list)
tri = {}
for line in reactions.splitlines():
    s, r = line.split(" => ")
    tr[s].append(r)
    tri[r] = s

seen = set()
for s, rs in tr.items():
    splitted = element.split(s)
    for i, (left, right) in enumerate(zip(splitted, splitted[1:])):
        for r in rs:
            new = splitted[:]
            new[i : i + 2] = [left + r + right]
            new = s.join(new)
            seen.add(new)
print("answer_a:", len(seen))

if "e => " in data:
    replacements = 0
    while element != "e":
        pos = {}
        for k, v in tri.items():
            delta = len(k) - len(v)
            if k in element:
                if v != "e" or len(element) - delta == 1:
                    pos[k] = (element.rfind(k) + len(k), delta)
        k = max(pos, key=pos.get)
        v = tri[k]
        # replace from right
        element = element[::-1].replace(k[::-1], v[::-1], 1)[::-1]
        replacements += 1
    print("answer_b:", replacements)



"""
--- Day 8: Two-Factor Authentication ---
https://adventofcode.com/2016/day/8
"""
import numpy as np
from aocd import data

from aoc_wim.ocr import AOCR


def animate(A, line):
    if line.startswith("rect"):
        w, h = [int(x) for x in line.split()[1].split("x")]
        A[0:h, 0:w] = "#"
    elif line.startswith("rotate"):
        i, shift = [int(x) for x in line.split("=")[1].split(" by ")]
        item = (i, slice(None)) if "row" in line else (slice(None), i)
        A[item] = np.roll(A[item], shift)
    return A


lines = data.splitlines()
im = (3, 7) if len(lines) == 4 else (6, 50)
A = np.full(im, ".")
for line in lines:
    A = animate(A, line)
print("answer_a:", (A == "#").sum())
print("answer_b:", AOCR[A])



"""
--- Day 23: A Long Walk ---
https://adventofcode.com/2023/day/23
"""
from aocd import data
from aoc_wim.zgrid import ZGrid


grid = ZGrid(data)
start = 1
end = grid.bottom_right - 1

nodes = {}
dzs = dict(zip("^>v<", [-1j, 1, 1j, -1]))
for z, g in grid.items():
    if g == "." and sum(grid.get(z + dz) in dzs for dz in dzs.values()) >= 2:
        nodes[z] = {}


def find_next_node(z0, dz0):
    path = [z0, z0 + dz0]
    while path[-1] not in nodes:
        [z1] = [z for z in grid.near(path[-1]) if z != path[-2] and grid.get(z, "#") != "#"]
        path.append(z1)
    return path[-1], path[-2] - path[-1], len(path) - 1


nodes[start] = {1j: find_next_node(start, 1j)}
nodes[end] = {-1j: find_next_node(end, -1j)}
for z0, branches in nodes.items():
    for dz0 in dzs.values():
        g = grid.get(z0 + dz0)
        if g in dzs and dz0 not in branches:
            z1, dz1, dist = find_next_node(z0, dz0)
            branches[dz0] = z1, dist
            nodes[z1][dz1] = z0, dist


def longest_path(part="a"):
    grid[start + 1j] = "v"
    paths = []
    stack = [(start, 0)]
    while stack:
        path = stack.pop()
        z0, d0 = path[-2:]
        if z0 == end:
            paths.append(path)
        branches = nodes[z0]
        for g, dz in dzs.items():
            if dz in branches and grid[z0 + dz] in ([g] if part == "a" else dzs):
                z1, dist = branches[dz]
                if z1 not in path:
                    stack.append(path + (z1, d0 + dist))
    return max(p[-1] for p in paths)


print("answer_a:", longest_path(part="a"))
print("answer_b:", longest_path(part="b"))



"""
--- Day 20: Firewall Rules ---
https://adventofcode.com/2016/day/20
"""
from aocd import data


intervals = []
for line in data.splitlines():
    left, right = line.split("-")
    interval = int(left), int(right)
    intervals.append(interval)
intervals.sort()

while True:
    n = len(intervals)
    for i in range(n - 1):
        lo_p, hi_p = intervals[i]
        lo, hi = intervals[i + 1]
        if lo <= hi_p + 1:
            del intervals[i + 1]
            intervals[i] = lo_p, max(hi_p, hi)
            break
    else:
        break

print("answer_a:", intervals[0][1] + 1)

b = 4294967296 if len(intervals) > 3 else 10
for lo, high in intervals:
    b -= high - lo + 1

print("answer_b:", b)



"""
--- Day 23: Unstable Diffusion ---
https://adventofcode.com/2022/day/23
"""
from collections import defaultdict
from collections import deque

from aocd import data

from aoc_wim.zgrid import ZGrid


grid = ZGrid(data, transform={".": None})
edges = deque([
    (-1j - 1, -1j, -1j + 1),
    (+1j - 1, +1j, +1j + 1),
    (-1j - 1,  -1, +1j - 1),
    (-1j + 1,  +1, +1j + 1),
])


def movers():
    d = defaultdict(list)  # dest: sources
    for z0 in grid:
        elves_nearby = {e[1]: sum(z0 + dz in grid for dz in e) for e in edges}
        if sum(elves_nearby.values()):
            for dz, n_nearby in elves_nearby.items():
                if not n_nearby:
                    d[z0 + dz].append(z0)
                    break
    return {z: zs[0] for z, zs in d.items() if len(zs) == 1}  # dest: source


# grid.draw(title="== Initial State == ")
i = 0
a = b = None
while a is None or b is None:
    i += 1
    z1_z0 = movers()
    grid.d = dict.fromkeys(grid.keys() - z1_z0.values() | z1_z0.keys(), "#")
    # grid.draw(title=f"== End of round {i} ==")
    edges.rotate(-1)
    if b is None and not z1_z0:  # nobody moved
        b = i
    if a is None and i == 10:
        a = grid.area - len(grid)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 7: The Treachery of Whales ---
https://adventofcode.com/2021/day/7
"""
import numpy as np
from aocd import data

A = np.array([int(n) for n in data.split(",")])
a = b = np.inf
for n in range(A.min(), A.max()):
    d = abs(A - n)
    a = min(a, d.sum())
    b = min(b, (d * (d + 1) // 2).sum())

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 8: Handheld Halting ---
https://adventofcode.com/2020/day/8
"""
from aocd import data

from aoc_wim.aoc2020 import Comp

# import logging; logging.basicConfig(format="%(message)s", level=logging.DEBUG)

ops = []
for line in data.splitlines():
    op, arg = line.split()
    ops.append((op, int(arg)))

comp = Comp(ops)
seen = {comp.line: comp.a}
while True:
    comp.step()
    if comp.line in seen:
        break
    seen[comp.line] = comp.a
print("answer_a:", comp.a)

comps = []
flipper = {"nop": "jmp", "jmp": "nop"}
for i, (op, arg) in enumerate(ops):
    if op in flipper:
        c_ops = ops[:]
        c_ops[i] = (flipper[op], arg)
        comp = Comp(c_ops)
        comps.append(comp)

done = False
while not done:
    for comp in comps:
        comp.step()
        if comp.line == len(ops):
            done = True
            break
print("answer_b:", comp.a)



"""
--- Day 18: Lavaduct Lagoon ---
https://adventofcode.com/2023/day/18
"""
from aocd import data

dzs = dict(zip("RDLU0123", (1, 1j, -1, -1j)*2))
za = zb = 0
zas = [za]
zbs = [zb]
for line in data.splitlines():
    d, n, c = line.replace("(#", "0x").rstrip(")").split()
    za += dzs[d] * int(n)
    zb += dzs[c[-1]] * int(c[:-1], 0)
    zas.append(za)
    zbs.append(zb)


def area(zs):
    # Shoelace formula + Pick's theorem
    a = p = 0
    for z0, z1 in zip(zs, zs[1:] + [zs[0]]):
        a += (z1.real + z0.real) * (z1.imag - z0.imag)
        p += abs(z1 - z0)
    return int(a/2 + p/2 + 1)


print("answer_a:", area(zas))
print("answer_b:", area(zbs))



"""
--- Day 3: Gear Ratios ---
https://adventofcode.com/2023/day/3
"""
from aocd import data

from aoc_wim.zgrid import ZGrid

g = ZGrid(data)
n = []
part = None
part_numbers = []
gear = []
gears = {z: [] for z in g.z("*", first=False)}
for z0, glyph in g.scan():
    if glyph.isdigit():
        n.append(glyph)
        for z in g.near(z0, n=8):
            c = g.get(z, ".")
            if not c.isdigit() and c != ".":
                part = True
                if c == "*":
                    gear.append(z)
        if not g.get(z0 + 1, ".").isdigit():
            if part:
                part_number = int("".join(n))
                part_numbers.append(part_number)
                if gear:
                    for z in set(gear):
                        gears[z].append(part_number)
                part = None
                gear.clear()
            n.clear()

print("answer_a:", sum(part_numbers))
print("answer_b:", sum([g[0] * g[1] for g in gears.values() if len(g) == 2]))



"""
--- Day 7: Camel Cards ---
https://adventofcode.com/2023/day/7
"""
from collections import Counter

from aocd import data


tr = str.maketrans("TQKA", "IKLM")


def key_a(h):
    return sorted(Counter(h[0]).values(), reverse=True), h[0].translate(tr)


def key_b(h):
    c = Counter(h[0].replace("J", "") or "J")
    k0, _ = key_a([h[0].replace("J", max(c, key=c.get))])
    return k0, h[0].translate(tr).replace("J", "0")


d = [h.split() for h in data.splitlines()]
print("answer_a", sum(i * int(b) for i, (h, b) in enumerate(sorted(d, key=key_a), 1)))
print("answer_b", sum(i * int(b) for i, (h, b) in enumerate(sorted(d, key=key_b), 1)))



import ast
import json
import pathlib

import pytest

from aoc_wim import plugin


here = pathlib.Path(__file__).parent
input_files = sorted(here.glob("20*/*/*.txt"))


def path2id(input_file):
    return str(input_file.relative_to(here))


def split_trailing_comments(lines):
    extra = []
    while lines and (not lines[-1].strip() or lines[-1].startswith("#")):
        extra.append(lines.pop())
    if len(lines) and "#" in lines[-1]:
        line, comment = lines[-1].split("#", 1)
        lines[-1] = line.strip()
        extra.append(comment.strip())
    if len(lines) > 1 and "#" in lines[-2]:
        line, comment = lines[-2].split("#", 1)
        lines[-2] = line.strip()
        extra.append(comment.strip())
    extra = [x.strip() for x in extra if x.strip()]
    return extra


def parse_extra_context(extra):
    result = {}
    for line in extra:
        equals = line.count("=")
        commas = line.count(",")
        if equals and equals == commas + 1:
            for part in line.split(","):
                k, v = part.strip().split("=")
                k = k.strip()
                v = v.strip()
                try:
                    v = ast.literal_eval(v)
                except ValueError:
                    pass
                if k in result:
                    raise NotImplementedError(f"Duplicate key {k!r}")
                result[k] = v
    return result


@pytest.mark.parametrize("input_file", input_files, ids=path2id)
def test_example(input_file, monkeypatch, request):
    # example input files are in ./YYYY/dd/fname.txt
    *pre, year, day, fname = input_file.parts
    year = int(year)
    day = int(day)
    # the head of each example file is an input data
    # the last two lines are part a and part b correct answers
    # there may be trailing comments after the answers
    lines = input_file.read_text(encoding="utf-8").splitlines()
    extra = split_trailing_comments(lines)
    extra = parse_extra_context(extra)
    if extra:
        monkeypatch.setenv("AOCD_EXTRA", json.dumps(extra))
    if len(lines) < 3:
        pytest.fail(f"test data {input_file} is malformed")
    *lines, expected_answer_a, expected_answer_b = lines
    input_data = "\n".join(lines).rstrip()
    # patch out aocd
    monkeypatch.setattr("aocd.data", input_data)
    # invoke the entrypoint with a controlled input
    actual_answer_a, actual_answer_b = plugin(year, day, input_data)
    # verify correct answers returned. sometimes a dash (-) may be used to indicate
    # an example input where the answer was provided for only one of the two parts
    if expected_answer_a != "-":
        assert actual_answer_a == expected_answer_a, f"{actual_answer_a=} {expected_answer_a=}"
    if expected_answer_b != "-" and not request.config.getoption("--part-a-only"):
        assert actual_answer_b == expected_answer_b, f"{actual_answer_b=} {expected_answer_b=}"



"""
--- Day 25: Clock Signal ---
https://adventofcode.com/2016/day/25
"""
import re

from aocd import data

x, y = re.findall(r"\d+", data)[:2]
a = 0b101010101010 - int(x) * int(y)
print("answer_a:", a)



"""
--- Day 23: Unstable Diffusion ---
https://adventofcode.com/2022/day/23
"""
import numpy as np
from aocd import data
from scipy.signal import convolve2d

from aoc_wim.ocr import autocrop


A = np.array([[int(c == "#") for c in line] for line in data.splitlines()])
mask = [1 << i for i in range(8)]
mask.insert(4, 0)
kernel = np.array(mask).reshape(3, 3)
maskN = kernel[-1,:].sum()
maskS = kernel[0,:].sum()
maskW = kernel[:,-1].sum()
maskE = kernel[:,0].sum()
masks = np.array([[maskN, maskS, maskW, maskE]]).T
offsets = np.array([
    [-1, 0],
    [1, 0],
    [0, -1],
    [0, 1],
])
a = b = None
r = 0
while a is None or b is None:
    r += 1
    A = np.pad(autocrop(A), pad_width=1)
    C = convolve2d(A, kernel, mode="same")
    w = np.argwhere(A)
    dirs = C[*w.T] & masks
    d0 = dirs == 0
    d = d0.argmax(axis=0)
    stay = d0.all(axis=0) | (~d0).all(axis=0)
    if b is None and stay.all():
        b = r
    w_ = w.copy()
    for i in range(4):
        w_[~stay & (d == i)] += offsets[i]
    yx, i, u = np.unique(w_, axis=0, return_counts=True, return_index=True)
    stay.fill(True)
    stay[i[u==1]] = False
    A.fill(0)
    A[*np.where(stay[:,None], w, w_).T] = 1
    offsets = np.roll(offsets, -1, axis=0)
    masks = np.roll(masks, -1, axis=0)
    if r == 10:
        a = (autocrop(A) == 0).sum()

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 15: Chiton ---
https://adventofcode.com/2021/day/15
"""
from aocd import data

from aoc_wim import zgrid
from aoc_wim.search import AStar


class Q15AStar(AStar):
    def adjacent(self, state):
        return [z for z in grid.near(state) if z in grid]

    def cost(self, current_state, next_state):
        return grid[next_state]

    def heuristic(self, current_state, next_state):
        return zgrid.manhattan_distance(current_state, next_state)


grid = zgrid.ZGrid(data, transform=int)
path = Q15AStar(state0=grid.top_left, target=grid.bottom_right).run()
print("answer_a:", sum(grid[z] for z in path) - grid[grid.top_left])

h, w = grid.height, grid.width
for z in zgrid.zrange(5 * (grid.bottom_right + 1 + 1j)):
    qr, rr = divmod(z.real, w)
    qi, ri = divmod(z.imag, h)
    grid[z] = (grid[complex(rr, ri)] + int(qr) + int(qi) - 1) % 9 + 1

path = Q15AStar(state0=grid.top_left, target=grid.bottom_right).run()
print("answer_b:", sum(grid[z] for z in path) - grid[grid.top_left])



"""
--- Day 5: How About a Nice Game of Chess? ---
https://adventofcode.com/2016/day/5
"""
import os
from collections import deque
from concurrent.futures import ProcessPoolExecutor
from itertools import count

from aocd import data

from aoc_wim.aoc2016 import md5_miner_q05


def main():
    code1 = []
    code2 = ["."] * 8
    remaining = set("01234567")
    chunksize = 50000
    blocks = ((data, i0, chunksize) for i0 in count(step=chunksize))
    with ProcessPoolExecutor() as pool:
        futures = deque([pool.submit(md5_miner_q05, *next(blocks)) for _ in range(os.cpu_count())])
        while True:
            for n, md5sum in futures.popleft().result():
                h5, h6 = md5sum[5], md5sum[6]
                code1.append(h5)
                if h5 in remaining:
                    code2[int(h5)] = h6
                    remaining.remove(h5)
            if not remaining:
                break
            futures.append(pool.submit(md5_miner_q05, *next(blocks)))

    print("answer_a:", "".join(code1[:8]))
    print("answer_b:", "".join(code2))


if __name__ == "__main__":
    main()



"""
--- Day 11: Hex Ed ---
https://adventofcode.com/2017/day/11
"""
from aocd import data

from aoc_wim import zgrid

z = d = dmax = 0
grid = zgrid.ZGrid({z: 1})
for step in data.split(","):
    z += zgrid.hexV[step]
    d = zgrid.hexagonal_distance(z)
    dmax = max(d, dmax)
    grid[z] = 1
    # grid.draw_hex(glyph=0, orientation="V", title=f" {d=}, {dmax=}")

print("answer_a:", d)
print("answer_b:", dmax)



"""
--- Day 17: Set and Forget ---
https://adventofcode.com/2019/day/17
"""
from aocd import data

from aoc_wim.aoc2019 import IntComputer
from aoc_wim.zgrid import ZGrid


def parsed(data):
    comp = IntComputer(data)
    comp.run()
    txt = "".join([chr(x) for x in reversed(comp.output)])
    grid = ZGrid(txt)
    grid.draw()
    return grid


def calibration(grid):
    result = 0
    for z0, txt in grid.items():
        if txt == "#" and {grid.get(z) for z in grid.near(z0)} == {"#"}:
            result += int(z0.real) * int(z0.imag)
    return result


def get_path(grid, compressed=True):
    # detect initial position and orientation
    [(z0, glyph)] = [(k,v) for (k,v) in grid.items() if v in "^>v<"]
    dz0 = grid.dzs[glyph]
    [z1] = [z for z in grid.near(z0) if grid.get(z) == "#"]
    steps = []

    # orient self onto scaffold
    if z0 + dz0 != z1:
        if z0 + dz0 * grid.turn_right == z1:
            steps.append("R")
            dz0 *= grid.turn_right
        elif z0 + dz0 * grid.turn_left == z1:
            steps.append("L")
            dz0 *= grid.turn_left
        else:
            assert z0 - dz0 == z1
            steps.extend("RR")
            dz0 *= grid.turn_around

    # find an uncompressed path
    z = z0
    dz = dz0
    while True:
        n = 0
        while grid.get(z + dz) == "#":
            n += 1
            z += dz
        steps.append(str(n))
        if grid.get(z + dz * grid.turn_right) == "#":
            dz *= grid.turn_right
            steps.append("R")
        elif grid.get(z + dz * grid.turn_left) == "#":
            dz *= grid.turn_left
            steps.append("L")
        else:
            break

    # compress path to memory requirement
    path = ",".join(steps)
    if compressed:
        options = compress(path)
        path = min(options, key=len)
        path += "\nn\n"  # suppress "continuous video feed"
    return path


def chunk_choices(path):
    choices = []
    for i in range(1, 20):
        chunk = path[:i]
        if path[i:i+1] not in {"", ","}:
            continue
        score = len(chunk) * path.count(chunk)
        choices.append((score, chunk))
    choices = {chunk: score for score, chunk in sorted(choices, reverse=True)}
    return choices


def compress(path, mem=20):
    results = []
    A_choices = chunk_choices(path)
    for A in A_choices:
        compressed_pathA = ""
        pathA = path
        while pathA.startswith(A):
            compressed_pathA += "A,"
            pathA = pathA[len(A):].lstrip(",")
        B_choices = chunk_choices(pathA)
        for B in B_choices:
            compressed_pathAB = compressed_pathA
            pathAB = pathA
            while pathAB.startswith((A, B)):
                if pathAB.startswith(A):
                    pathAB = pathAB[len(A):].lstrip(",")
                    compressed_pathAB += "A,"
                if pathAB.startswith(B):
                    pathAB = pathAB[len(B):].lstrip(",")
                    compressed_pathAB += "B,"
            C_choices = chunk_choices(pathAB)
            for C in C_choices:
                compressed_pathABC = compressed_pathAB
                pathABC = pathAB
                while pathABC.startswith((A, B, C)):
                    if pathABC.startswith(A):
                        pathABC = pathABC[len(A):].lstrip(",")
                        compressed_pathABC += "A,"
                    if pathABC.startswith(B):
                        pathABC = pathABC[len(B):].lstrip(",")
                        compressed_pathABC += "B,"
                    if pathABC.startswith(C):
                        pathABC = pathABC[len(C):].lstrip(",")
                        compressed_pathABC += "C,"
                if not pathABC:
                    result = compressed_pathABC.rstrip(",")
                    if len(result) <= mem:
                        results.append("\n".join([result, A, B, C]))
    return results


if __name__ == "__main__":
    grid = parsed(data)
    print("answer_a:", calibration(grid))

    comp = IntComputer(data)
    comp.reg[0] = 2
    path = get_path(grid)
    comp.input_text(path)
    comp.run()
    print("answer_b:", comp.output[0])



"""
--- Day 10: Syntax Scoring ---
https://adventofcode.com/2021/day/10
"""
from statistics import median

from aocd import data


class ParserError(Exception):
    def __init__(self, line, expected, found):
        self.line = line
        self.expected = expected
        self.found = found
        self.syntax_error_score = [3, 57, 1197, 25137][")]}>".index(found)]

    def __str__(self):
        return f"{self.line} - Expected {self.expected}, but found {self.found} instead."


class Parser:

    lookup = dict(zip("([{<", ")]}>"))

    def __init__(self):
        self.pos = 0
        self.stack = []

    def __call__(self, line):
        while self.pos < len(line):
            found = line[self.pos]
            if found in self.lookup:
                self.stack.append(found)
            else:
                expected = self.lookup[self.stack[-1]]
                if found != expected:
                    raise ParserError(line, expected, found)
                self.stack.pop()
            self.pos += 1

    @property
    def autocompletion(self):
        return "".join(self.lookup[char] for char in reversed(self.stack))

    @property
    def autocomplete_score(self):
        result = 0
        for char in self.autocompletion:
            result *= 5
            result += ")]}>".index(char) + 1
        return result


a = b = 0
bs = []
for line in data.splitlines():
    parser = Parser()
    try:
        parser(line)
    except ParserError as err:
        a += err.syntax_error_score
    else:
        bs.append(parser.autocomplete_score)

if bs:
    b = median(bs)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 14: Parabolic Reflector Dish ---
https://adventofcode.com/2023/day/14
"""
from aocd import data


def tilt_north(data):
    rows = data.splitlines()
    cols = [''.join(col) for col in list(zip(*rows))]
    newcols = []
    for col in cols:
        segments = col.split("#")
        segments = [''.join(sorted(s, reverse=True)) for s in segments]
        newcol = "#".join(segments)
        newcols.append(newcol)
    newrows = [''.join(row) for row in list(zip(*newcols))]
    newdata = "\n".join(newrows)
    return newdata


def rotate_right(data):
    rows = data.splitlines()
    cols = [''.join(col) for col in list(zip(*rows))]
    return "\n".join([col[::-1] for col in cols])


def one_cycle(data):
    for _ in "NWSE":
        data = tilt_north(data)
        data = rotate_right(data)
    return data


def load_on_north_support_beams(rows):
    return sum(i*row.count("O") for i, row in enumerate(rows.splitlines()[::-1], 1))


print("answer_a:", load_on_north_support_beams(tilt_north(data)))
cycles = 1_000_000_000
seen = {}
period = None
i = 0
while i < cycles:
    data = one_cycle(data)
    i += 1
    # print(f"After {i+1} cycle:\n{data}\n\n")
    if period is None and data in seen:
        period = i - seen[data]
        i += period * ((cycles - i) // period)
    seen[data] = i
print("answer_b:", load_on_north_support_beams(data))



import logging
from collections import deque


log = logging.getLogger(__name__)


class Comp:

    def __init__(self, data, pid=None, ch_snd=None, ch_rcv=None, a0=0):
        self.code = [x.split() for x in data.splitlines()]
        self.reg = {r: 0 for op, r, *rest in self.code}
        self.reg.pop("1", None)
        self.reg = dict(sorted(self.reg.items()))
        self.pid = pid
        if self.pid is not None:
            self.reg["p"] = pid
        self.reg["a"] = a0
        if ch_snd is None:
            ch_snd = deque()
        if ch_rcv is None:
            ch_rcv = deque()
        self.ch_snd = ch_snd
        self.ch_rcv = ch_rcv
        self.line = 0
        self.n = 0
        self.n_snd = 0
        self.n_mul = 0

    def snd(self, x):
        if x in self.reg or self.pid is not None:
            self.n_snd += 1
            try:
                self.ch_snd.append(self.reg[x])
            except KeyError:
                return

    def rcv(self, x):
        if self.reg[x] or self.pid is not None:
            self.reg[x] = self.ch_rcv.popleft()

    def set(self, x, y):
        self.reg[x] = y

    def add(self, x, y):
        self.reg[x] += y

    def mul(self, x, y):
        self.reg[x] *= y
        self.n_mul += 1

    def mod(self, x, y):
        self.reg[x] %= y

    def jgz(self, x, y):
        x = self.reg[x] if x in self.reg else int(x)
        if x > 0:
            self.line += y - 1

    def sub(self, x, y):
        self.reg[x] -= y

    def jnz(self, x, y):
        x = self.reg[x] if x in self.reg else int(x)
        if x:
            self.line += y - 1

    def wtf(self, x, y):
        self.reg[x] = all(y % i for i in range(2, y))

    def blocked(self):
        if not (0 <= self.line < len(self.code)):
            return True
        if self.code[self.line][0] == "rcv" and not self.ch_rcv:
            return True

    def step(self):
        self.n += 1
        op, r, *args = self.code[self.line]
        op_func = getattr(self, op)
        if args:
            [y] = args
            y = self.reg[y] if y in self.reg else int(y)
            args = y,
        else:
            y = "."
        log.debug(f"{self.pid} {self.n:5}:   L{self.line:03d} {op:3} {r:1} {y:14}   snd={self.ch_snd[-1] if self.ch_snd else '':4}   rcv={self.ch_rcv[-1] if self.ch_rcv else '':4}  {self.reg}  n_mul={self.n_mul}")
        op_func(r, *args)
        self.line += 1



"""
--- Day 15: Beverage Bandits ---
https://adventofcode.com/2018/day/15
"""
from collections import deque
from operator import attrgetter
from time import sleep

from aocd import data
from termcolor import colored


# TODO: refactor to use zgrid


class CombatEnds(Exception):
    pass


class Unit:
    grid = None
    ap = None
    glyph = "U"

    def __init__(self, pos, id):
        self.id = id
        self.pos = pos
        self.hp = 200

    @property
    def yx(self):
        return int(self.pos.imag), int(self.pos.real)

    @property
    def alive(self):
        return self.hp > 0

    @property
    def targets(self):
        units = self.grid["units"]
        result = [u for u in units if u.alive and u.glyph != self.glyph]
        if not result:
            raise CombatEnds
        return result

    @property
    def targets_adjacent(self):
        return [u for u in self.targets if abs(u.pos - self.pos) == 1]

    @property
    def targets_with_available_squares(self):
        return [u for u in self.targets if u.available_squares]

    @property
    def available_squares(self):
        units = self.grid["units"]
        candidates = [self.pos - 1j, self.pos - 1, self.pos + 1, self.pos + 1j]
        not_wall = [c for c in candidates if self.grid.get(c, "#") != "#"]
        unit_positions = {u.pos for u in units if u.alive}
        avail = [c for c in not_wall if c not in unit_positions]
        return avail

    def attack(self, other):
        assert abs(self.pos - other.pos) == 1
        assert other.alive
        other.hp -= self.ap

    def move(self):
        new_pos_choices = self.available_squares
        assert new_pos_choices
        ops = self.targets_with_available_squares
        if not ops:
            return
        target_positions = {pos for o in ops for pos in o.available_squares}
        assert target_positions
        units = self.grid["units"]
        others_positions = {o.pos for o in units if o.alive and o is not self}
        state0 = self.pos
        found = bfs(
            state0,
            targets=target_positions,
            grid=self.grid,
            others_positions=others_positions,
        )
        if not found:
            return
        [shortest_path] = {d for (pos, d) in found}
        target_positions = [pos for (pos, d) in found]
        target_positions.sort(key=lambda pos: (pos.imag, pos.real))
        target_pos = target_positions[0]
        distances = {}
        for choice in new_pos_choices:
            found = bfs(
                choice,
                targets={target_pos},
                grid=self.grid,
                others_positions=others_positions,
            )
            [(t, distances[choice])] = found
            assert t == target_pos
        min_dist = min(distances.values())
        assert min_dist == shortest_path - 1
        choices = [c for (c, d) in distances.items() if d == min_dist]
        choices.sort(key=lambda pos: (pos.imag, pos.real))
        new_pos = choices[0]
        assert abs(self.pos - new_pos) == 1
        self.pos = new_pos

    @property
    def opponent(self):
        ops = self.targets_adjacent
        if ops:
            min_hp = min([o.hp for o in ops])
            ops = [o for o in ops if o.hp == min_hp]
            ops.sort(key=attrgetter("yx"))
            return ops[0]

    def act(self):
        op = self.opponent
        if op is not None:
            self.attack(other=op)
            return
        places = self.available_squares
        if not places:
            return
        self.move()
        op = self.opponent
        if op is not None:
            self.attack(other=op)
            return


class Elf(Unit):
    glyph = "E"
    cglyph = colored(glyph, "green", attrs=["bold"])
    ap = 3


class Goblin(Unit):
    glyph = "G"
    cglyph = colored(glyph, "red", attrs=["bold"])
    ap = 3


def draw(grid, round=0, interactive=False, dt=0.1):
    print("\33c")
    units = grid["units"]
    units_on_grid = {}
    for unit in units:
        if unit.alive:
            units_on_grid[unit.pos] = unit
    print()
    if round:
        print(f"After {round} round{'s' if round > 1 else ''}:")
    else:
        print("Initially:")
    for y in grid["y-axis"]:
        line = []
        units_on_line = []
        for x in grid["x-axis"]:
            pos = complex(x, y)
            if pos in units_on_grid:
                unit = units_on_grid[pos]
                line.append(unit.cglyph)
                units_on_line.append(unit)
            else:
                line.append(grid[pos])
        print(*line, sep="", end="")
        if units_on_line:
            units_str = ", ".join([f"{u.cglyph}({u.hp})" for u in units_on_line])
            print("   " + units_str)
        else:
            print()
    if interactive:
        input("Press enter to continue ...")
    else:
        sleep(dt)


def parsed(data):
    grid = {}
    lines = data.splitlines()
    h = len(lines)
    [w] = {len(x) for x in lines}
    grid["y-axis"] = range(h)
    grid["x-axis"] = range(w)
    uid = 0
    units = []
    for y, line in enumerate(lines):
        for x, char in enumerate(line):
            pos = complex(x, y)
            if char in "#.":
                grid[pos] = char
            elif char == "G":
                g = Goblin(id=uid, pos=pos)
                units.append(g)
                grid[pos] = "."
                uid += 1
            elif char == "E":
                e = Elf(id=uid, pos=pos)
                units.append(e)
                grid[pos] = "."
                uid += 1
    grid["units"] = units
    return grid


def bfs(state0, targets, grid, others_positions):
    depth = 0
    queue = deque([(state0, depth)])
    seen = {state0}
    i = 0
    found = set()
    shortest_path = None
    while queue:
        state, new_depth = queue.popleft()
        i += 1
        if new_depth > depth:
            depth = new_depth
            if shortest_path is not None and depth > shortest_path:
                assert found
                return found
        if state in targets:
            if shortest_path is not None:
                assert shortest_path == depth
            shortest_path = depth
            found.add((state, depth))
        candidates = [state - 1j, state - 1, state + 1, state + 1j]
        ok = [
            c
            for c in candidates
            if grid.get(c, "#") != "#" and c not in others_positions and c not in seen
        ]
        queue.extend((child, depth + 1) for child in ok)
        seen.update(ok)
    return found


def part_a(data):
    Unit.grid = grid = parsed(data)
    round = 0
    units = grid["units"]
    n_elves = sum(1 for u in units if isinstance(u, Elf))
    while True:
        # draw(grid, round)
        units.sort(key=attrgetter("yx"))
        for unit in units:
            if unit.alive:
                try:
                    unit.act()
                except CombatEnds:
                    score = sum(u.hp for u in units if u.alive)
                    elves_alive = [u for u in units if isinstance(u, Elf) and u.alive]
                    part_a.all_elves_alive = n_elves == len(elves_alive)
                    return round * score
        round += 1


def part_b(data):
    Elf.ap = 3
    while True:
        Elf.ap += 1
        outcome = part_a(data)
        if part_a.all_elves_alive:
            Elf.ap = 3
            return outcome


print("answer_a:", part_a(data))
print("answer_b:", part_b(data))



from aocd import data


def safe(ns0):
    ns = sorted(ns0)
    if ns != ns0 and ns != ns0[::-1]:
        return False
    return all(1 <= j - i <= 3 for i, j in zip(ns, ns[1:]))


a = b = 0
for line in data.splitlines():
    levels = [int(n) for n in line.split()]
    if safe(levels):
        a += 1
        b += 1
        continue
    for i in range(len(levels)):
        b_levels = levels.copy()
        del b_levels[i]
        if safe(b_levels):
            b += 1
            break

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 24: Immune System Simulator 20XX ---
https://adventofcode.com/2018/day/24
"""
import logging
from operator import attrgetter

from aocd import data
from parse import parse


log = logging.getLogger(__name__)
# logging.basicConfig(format="%(message)s", level=logging.DEBUG)


class StaleMate(Exception):
    """Neither infection nor immune can deal damage to eachother"""


class System:
    def __init__(self, text, boost=0):
        self.name, text = text.split(":\n")
        text = text.replace("\n ", " ")
        self.groups = []
        for i, line in enumerate(text.splitlines(), start=1):
            self.groups.append(Group(id=i, text=line, system=self, boost=boost))

    def select_targets_from(self, other_system):
        taken = set()
        key = attrgetter("power", "initiative")
        groups = sorted(self.groups, key=key, reverse=True)
        groups = [g for g in groups if g.alive]
        for group in groups:
            other = [g for g in other_system.groups if g.id not in taken and g.alive]
            group.select_target_from(other)
            if group.target is not None:
                assert group.target not in taken, "group ids must be unique"
                taken.add(group.target.id)

    @property
    def alive(self):
        return any(g.alive for g in self.groups)

    @property
    def n(self):
        return sum(g.n for g in self.groups)


class Group:

    template = "{:d} units each with {:d} hit points{}with an attack that does {:d} {} damage at initiative {:d}"

    def __init__(self, id, text, system, boost=0):
        self.id = id
        self.system = system
        parsed = parse(self.template, text).fixed
        self.n, self.hp, props, self.ap, self.atype, self.initiative = parsed
        self.ap += boost
        self.weak = []
        self.immune = []
        for prop in props.strip(" ()").split("; "):
            if prop.startswith("immune to "):
                self.immune += prop.removeprefix("immune to ").split(", ")
            elif prop.startswith("weak to "):
                self.weak += prop.removeprefix("weak to ").split(", ")
        self.target = None

    @property
    def alive(self):
        return self.n > 0

    @property
    def power(self):
        return self.n * self.ap

    def attack(self):
        assert self.target is not None
        assert self.target.alive
        kills = self.damage(self.target) // self.target.hp
        kills = min(kills, self.target.n)
        self.target.n -= kills
        log.debug(
            "%s group %d attacks defending group %d, killing %d units",
            self.system.name,
            self.id,
            self.target.id,
            kills,
        )

    def damage(self, other_group):
        multiplier = 1
        if self.atype in other_group.immune:
            multiplier = 0
        elif self.atype in other_group.weak:
            multiplier = 2
        return self.power * multiplier

    def select_target_from(self, other_groups):
        targets = []
        for other_group in other_groups:
            damage = self.damage(other_group)
            if damage > 0:
                log.debug(
                    "%s group %d would deal defending group %d %d damage",
                    self.system.name,
                    self.id,
                    other_group.id,
                    damage,
                )
                targets.append(other_group)
        if not targets:
            self.target = None
            return
        self.target = max(
            targets, key=lambda t: (self.damage(t), t.power, t.initiative)
        )


class Battle:

    def __init__(self, data, immuno_boost=0):
        self.ticks = 0
        immune, infection = data.split("\n\n")
        self.immune = System(immune, boost=immuno_boost)
        self.infection = System(infection)

    def target_selection_phase(self):
        self.infection.select_targets_from(self.immune)
        self.immune.select_targets_from(self.infection)

    def attacking_phase(self):
        all_groups = self.infection.groups + self.immune.groups
        # sort in attacking order
        all_groups.sort(key=attrgetter("initiative"), reverse=True)
        for group in all_groups:
            if group.target is not None and group.target.alive:
                group.attack()

    def tick(self):
        for system in self.immune, self.infection:
            log.debug(system.name + ":")
            if system.alive:
                for group in system.groups:
                    if group.alive:
                        log.debug("Group %d contains %d units", group.id, group.n)
            else:
                log.debug("No groups remain.")
        log.debug("")
        self.target_selection_phase()
        log.debug("")
        n_before = self.immune.n + self.infection.n
        self.attacking_phase()
        n_after = self.immune.n + self.infection.n
        if n_after == n_before:
            raise StaleMate
        log.debug("")
        self.ticks += 1

    def run_until_complete(self):
        while self.infection.alive and self.immune.alive:
            self.tick()

    def part_a(self):
        self.run_until_complete()
        if self.immune.alive:
            assert not self.infection.alive
            winner = self.immune
        else:
            assert self.infection.alive
            winner = self.infection
        return winner.n


print("answer_a:", Battle(data).part_a())

boost = 0
while True:
    battle = Battle(data, immuno_boost=boost)
    try:
        battle.run_until_complete()
    except StaleMate:
        pass
    if battle.immune.alive and not battle.infection.alive:
        break
    boost += 1

print("answer_b:", battle.immune.n)



"""
--- Day 11: Cosmic Expansion ---
https://adventofcode.com/2023/day/11
"""
from itertools import combinations

from aocd import data

from aoc_wim.zgrid import manhattan_distance
from aoc_wim.zgrid import ZGrid

grid = ZGrid(data)
h, w = grid.height, grid.width
gs = grid.z("#", first=False)
g_cols = {g.real for g in gs}
g_rows = {g.imag for g in gs}
empty_cols = [i for i in range(w) if i not in g_cols]
empty_rows = [i for i in range(h) if i not in g_rows]

# expansion factor for part b
f = 1_000_000
if len(gs) == 9:
    # 9 galaxies for the example data
    # use a smaller expansion factor
    f = 100

a = b = 0
for g0, g1 in combinations(gs, 2):
    d = manhattan_distance(g0, g1)
    c01 = range(*sorted([int(g0.real), int(g1.real)]))
    r01 = range(*sorted([int(g0.imag), int(g1.imag)]))
    c = sum(1 for c in empty_cols if c in c01)
    r = sum(1 for r in empty_rows if r in r01)
    a += d + c + r
    b += d + (c + r) * (f - 1)

print("answer_a:", a)
print("answer_b:", b)



"""
--- Day 2: I Was Told There Would Be No Math ---
https://adventofcode.com/2015/day/2
"""
from aocd import data


def box_area(w, h, l):
    return 2 * l * w + 2 * w * h + 2 * h * l


def smallest_side(w, h, l):
    return min(l * w, w * h, h * l)


def shortest_perimeter(w, h, l):
    return 2 * min(l + w, w + h, h + l)


def box_volume(w, h, l):
    return w * h * l


area = 0
length = 0
for line in data.splitlines():
    w, h, l = [int(d) for d in line.split("x")]
    area += box_area(w, h, l) + smallest_side(w, h, l)
    length += shortest_perimeter(w, h, l) + box_volume(w, h, l)


print("answer_a:", area)
print("answer_b:", length)



from collections import Counter
from functools import lru_cache
from math import sqrt


def rsubset_sum(vals, target=0):
    # recursion impl

    @lru_cache(maxsize=None)
    def worker(hand=frozenset()):
        sum_ = sum(vals[i] for i in hand)
        if sum_ == target:
            result = {hand}
        elif sum_ > target:
            result = set()
        else:
            choices = set(range(len(vals))) - hand
            hands = (worker(hand | {choice}) for choice in choices)
            result = set().union(*hands)
        return result

    result = [[vals[i] for i in way] for way in set(worker())]
    return result


def subset_sum(vals, target=0):
    # dynamic programming impl
    sums = {0: [()]}  # key=sum, value=list of subsets for the sum
    if target in sums:
        yield from sums[target]  # annoying base case
    for val in vals:
        items = sums.items()  # don't change dict size during iteration
        sums = dict(items)
        for prev_sum, prev_subsets in items:
            sum_ = prev_sum + val
            subsets = [s + (val,) for s in prev_subsets]
            sums[sum_] = sums.get(sum_, []) + subsets
            if sum_ == target:
                yield from subsets


def ways(total, coins=(1, 2, 5, 10, 20, 50, 100)):
    ways = [[Counter()]] + [[] for _ in range(total)]
    for coin in coins:
        for i in range(coin, total + 1):
            ways[i] += [way + Counter({coin: 1}) for way in ways[i - coin]]
    return ways[total]


class Tribonacci(dict):
    """https://oeis.org/A000073"""

    def __init__(self):
        self[0] = self[1] = 0
        self[2] = 1

    def __missing__(self, key):
        val = self[key - 1] + self[key - 2] + self[key - 3]
        self[key] = val
        return val


def factorise(n):
    """returns the sorted prime factors of natural number n

    >>> factorise(12)
    [2, 2, 3]
    >>> factorise(4998)
    [2, 3, 7, 7, 17]
    >>> factorise(4999)
    [4999]"""
    if n < 0:
        result = factorise(-n)
        result[0] *= -1
        return result
    for i in range(2, int(1 + sqrt(n))):
        if n % i == 0:
            return [i] + factorise(n // i)
    else:
        return [n]


def extended_euclidean(a, b):
    # find m1, m2 such that a*m1 + b*m2 = gcd(a, b)
    r_prev, r = a, b
    s_prev, s = 1, 0
    t_prev, t = 0, 1
    while r:
        q = r_prev // r
        r_prev, r = r, r_prev - q * r
        s_prev, s = s, s_prev - q * s
        t_prev, t = t, t_prev - q * t
    if s_prev < 0:
        s_prev += a
    if t_prev < 0:
        t_prev += b
    return s_prev, t_prev



"""
--- Day 21: Springdroid Adventure ---
https://adventofcode.com/2019/day/21
"""
from aocd import data

from aoc_wim.aoc2019 import IntComputer

# Jump if (there's a hole 1 space away (not A)
#       or there's a hole 2 spaces away (not B)
#       or there's a hole 3 spaces away (not C))
# and there's ground 4 spaces away (D)
prog_a = """\
NOT A J
NOT B T
OR T J
NOT C T
OR T J
AND D J
WALK
"""

# Jump if (there's a hole 1 space away (not A)
#       or there's a hole 2 spaces away (not B)
#       or there's a hole 3 spaces away (not C))
# and there's ground 4 spaces away (NOT NOT D)
# and (there's ground 8 spaces away (H)
#      or 5 spaces away (E))
prog_b = """\
NOT A J
NOT B T
OR T J
NOT C T
OR T J
NOT D T
NOT T T
AND T J
AND H T
OR E T
AND T J
RUN
"""

for part, prog in zip("ab", [prog_a, prog_b]):
    comp = IntComputer(data, inputs=[ord(x) for x in reversed(prog)])
    comp.run()
    val = comp.output.popleft()
    print(*[chr(x) for x in reversed(comp.output)], sep="")
    print(f"answer_{part}:", val)



"""
--- Day 14: Docking Data ---
https://adventofcode.com/2020/day/14
"""
from itertools import combinations

from aocd import data

da = {}
db = {}
for line in data.splitlines():
    if line.startswith("mask"):
        mask = line.split()[-1]
        on = int(mask.replace("X", "0"), 2)
        off = int(mask.replace("X", "1"), 2)
        Xi = [1 << i for i, v in enumerate(reversed(mask)) if v == "X"]
        nX = mask.count("X")
    else:
        mem, val = line.split(" = ")
        mem = int(mem.strip("mem[]"))
        val = int(val)
        da[mem] = (val | on) & off
        if nX > 10:
            # hackish: skip an extremely slow part b for the test of part a
            continue
        for r in range(nX + 1):
            for comb in combinations(Xi, r):
                memb = mem | on
                for bit in comb:
                    memb ^= bit
                db[memb] = val

print("answer_a:", sum(da.values()))
print("answer_b:", sum(db.values()))



"""
--- Day 8: Matchsticks ---
https://adventofcode.com/2015/day/8
"""
from aocd import data


def tokens(s):
    iterator = iter(s)
    for char in iterator:
        if char == "\\":
            char += next(iterator)
            if char.endswith("x"):
                char += next(iterator)
                char += next(iterator)
        yield char


def encoder(s):
    iterator = iter(s)
    yield '"'
    for char in iterator:
        if char == '"' or char == "\\":
            yield "\\"
        yield char
    yield '"'


def tokens_len(s):
    return sum(1 for token in tokens(s)) - 2


def length_diff(data):
    return sum(len(line) - tokens_len(line) for line in data.splitlines())


def encoded_diff(data):
    return sum(len("".join(encoder(line))) - len(line) for line in data.splitlines())


print("answer_a:", length_diff(data))
print("answer_b:", encoded_diff(data))



"""Wim's solutions for https://adventofcode.com/"""
import io
import runpy
import sys
from pathlib import Path


def dynamic_version():
    here = Path(__file__).parent
    years = here.glob("aoc20??")
    max_year = max(years)
    days = max_year.glob("q??.py")
    max_day = max(days)
    year = int(max_year.name[-4:])
    day = int(max_day.name[1:3])
    return f"{year}.{day}"


__version__ = dynamic_version()


def plugin(year, day, data):
    mod_name = "aoc_wim.aoc{}.q{:02d}".format(year, day)
    sys.modules.pop(mod_name, None)
    old_stdout = sys.stdout
    sys.stdout = out = io.StringIO()
    try:
        import aocd
        aocd.data = data
        runpy.run_module(mod_name, run_name="__main__")
    finally:
        sys.stdout = old_stdout
        del aocd.data
    lines = [x for x in out.getvalue().splitlines() if x]
    answer_a = answer_b = None
    for line in lines:
        if line.startswith("answer_a"):
            if len(line.split()) > 1:
                answer_a = line.split()[-1]
            else:
                answer_a = ""
        elif line.startswith("answer_b"):
            if len(line.split()) > 1:
                answer_b = line.split()[-1]
            else:
                answer_b = ""
    if answer_a is not None and answer_b is not None:
        return answer_a, answer_b
    if not lines:
        return None, None
    if len(lines) == 1:
        answer_a = lines[0].split()[-1]
    else:
        if answer_a is None:
            answer_a = lines[-2].split()[-1]
        if answer_b is None:
            answer_b = lines[-1].split()[-1]
    return answer_a, answer_b



"""
--- Day 21: Scrambled Letters and Hash ---
https://adventofcode.com/2016/day/21
"""
from itertools import permutations

from aocd import data


def swap_position(list_, words):
    """swap position x with position y"""
    x, y = int(words[0]), int(words[-1])
    list_[x], list_[y] = list_[y], list_[x]


def swap_letter(list_, words):
    """swap letter x with letter y"""
    x, y = words[0], words[-1]
    tr = {x: y, y: x}
    list_[:] = [tr.get(k, k) for k in list_]


def rotate_left(list_, words):
    """rotate left x steps"""
    x = int(words[0])
    list_[:] = list_[x:] + list_[:x]


def rotate_right(list_, words):
    """rotate right x steps"""
    x = int(words[0])
    x = len(list_) - x
    list_[:] = list_[x:] + list_[:x]


def rotate_based(list_, words):
    """rotate based on position of letter x"""
    x = words[-1]
    n = list_.index(x)
    if n >= 4:
        n += 1
    n += 1
    n = len(list_) - n
    list_[:] = list_[n:] + list_[:n]


def reverse_positions(list_, words):
    """reverse positions 0 through 4"""
    x, y = int(words[0]), int(words[-1])
    list_[x : y + 1] = list_[x : y + 1][::-1]


def move_position(list_, words):
    """move position x to position y"""
    x, y = int(words[0]), int(words[-1])
    if x < y:
        list_.insert(y + 1, list_[x])
        del list_[x]
    elif y < x:
        list_.insert(y, list_[x])
        del list_[x + 1]


def scramble(data, original):
    list_ = list(original)
    for line in data.splitlines():
        line = line.replace(" ", "_", 1)
        func_name, *words = line.split()
        globals()[func_name](list_, words)
    return "".join(list_)


def descramble(data, scrambled):
    for original in permutations(scrambled):
        if scramble(data, original) == scrambled:
            return "".join(original)


start = "abcde" if data.count("\n") < 10 else "abcdefgh"
print("answer_a:", scramble(data, start))
print("answer_b:", descramble(data, "fbgdceah"))
